# # #
# /5/env.rb
# (c) 2009 Jan Lelis <mail@janlelis.de>. MIT License.
# See: https://rbjl.janlelis.com/5-playing-with-dijkstra
# # #

require 'set' # std lib for sets [alternative to arrays]

# Env = Environment = labyrinth - loads the labyrinth file and do things with it
class Env
  attr_reader :array, :nodes, :edges, :path # getter for instance vars

  # reads file into lab-array
  def initialize(prm_path)
    @array = []

    raise ArgumentError, 'Sorry, could not load the labyrinth file "'<< File.full_path(prm_path,true) << '"' unless File.exist?(File.full_path(prm_path))

    File.open(File.full_path(prm_path)) do |b_file| # Read file..
      b_file.each {|b_line| b_line.rstrip!; @array << b_line.split(';').collect {|a_ele| a_ele.to_i} } # ..and put ele as int into an multi-dim-array [A and Z are converted to 0 by to_i]
    end

    raise ArgumentError, 'Sorry, could not load the labyrinth: the labyrinth file is empty!' if @array.empty? || @array[0].empty?
    raise ArgumentError, 'Sorry, could not load the labyrinth: wrong format!' unless @array.size == 9 # only a weak test
  end


  # get the solution :)
  def output_path(prm_algo=:dijkstra,prm_format=:console)
    # choose algo
    algo = (prm_algo==:backtracking) ? Algo.const_get(prm_algo.to_s.capitalize) : Algo::Dijkstra
    # (currently, the std algo is Dijkstra -- needs to be changed if algos get more complex or more are added!)

    # init calculation
    transform_to_graph if algo::NeedsGraph

    case prm_algo
      when :dijkstra
        algo.init(@nodes,@edges)
      when :astar
        # with a block appended describing the heuristic, the dijkstra module turns into A* :]
        algo.init(@nodes,@edges) do |idx,tupel|
          tupel.d + 16-(idx/9+idx%9) # ...
        end
      when :astar_b
        h = [16, 15, 14, 13, 12, 11, 10, 9, 8, 15, 14, 13, 12, 11, 10, 9, 8, 7, 14, 13, 12, 11, 10, 9, 8, 7, 6, 13, 12, 11, 10, 9, 8, 7, 6, 5, 12, 11, 10, 9, 8, 7, 6, 5, 4, 11, 10, 9,8, 7, 6, 5, 4, 3, 10, 9, 8, 7, 6, 5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2, 1, 8, 7, 6, 5, 4, 3, 2, 1, 0]
        algo.init(@nodes,@edges) do |idx,tupel|
          tupel.d + h[idx] # same as above, only with hardcoded values
        end
      when :backtracking
        algo.init(@array)
    end

    # calculate! :)
    @path = algo.calc

    # put out
    if prm_format == :console or prm_format == :both
      puts "Algorithm:       #{prm_algo.capitalize}"
      puts '_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _'
      if @path and @path.is_a? Array
        puts algo::info
        puts 'Path:'
      end
      puts @path
    end

    # html output in a very ugly way ;)
    if prm_format == :html or prm_format == :both
      # calculate file name
      now_name = "#{Config::BasePathHtml}#{Time.now.strftime("%y%m%d_%H%M%S_" + Config[:file].to_s + "_" + (Config[:algo]||:dijkstra).to_s)}.htm"
      # and create sub dir if not present
      Dir.mkdir 'html' unless File.directory? 'html'

      # write the file ;)
      File.open(now_name,'w') do |out|
        out.puts "<html><head><title>AKMRubyLabyrinth - File: #{Config[:file]} - Algorithm: #{(Config[:algo]||:dijkstra).to_s.capitalize}</title>"
        out.puts '<style type="text/css">
 body 	              {background:#333;font-family:Arial;color:white}
 h1,h3,.info          {background:#777;border:darkred solid 1px}
 h1,h2,h3,.info,.sub  {padding:5px;margin:5px}
 .sub                 {background:#777;border:darkred dashed 1px}
</style></head><body>'# some style blabla
        out.puts '<h1>AKMRubyLabyrinth</h1>'
        out.puts "<h3>File: #{Config[:file]}<br/>"
        out.puts "Algorithm: #{(Config[:algo]||:dijkstra).to_s.capitalize}</h3>"
        out.puts '<div class="info">'
        if path.is_a? Array
          out.puts " <div class=\"sub\">Path length: #{path.size}</div>"
          out.puts " <div class=\"info\">Solution: <strong>"
          path.each {|step| out.print " #{step}"}
        elsif
          out.puts "<div class=\"sub\"><strong>#{path}"
        end
        out.puts ' </strong></div>'
        out.puts '</div>'

        out.puts '</body></html>'
      end
      # (weak) check if it was successful
      if File.file? now_name
        puts 'Output saved in ' << now_name
      else
        puts 'Sorry, could not save the solution in a html document..'
      end
    end

    # or clean °_°
    if prm_format == :clean
      puts @path
    end
  end


  # transforms @array into @nodes and @edges
  def transform_to_graph()
    @edges = Set.new  # start
    @nodes = Set.new  #      values

    9.times do |i| # 0.upto 9
	    9.times do |j|
        if @array[i][j] == 0 # field is valid node
          ni = node_ident(i,j)
          @nodes << ni
          @edges << [node_ident(i,j-1),ni] << [ni,node_ident(i,j-1)] if j>0 and @array[i][j-1] == 0 # valid edge version 1
          @edges << [node_ident(i-1,j),ni] << [ni,node_ident(i-1,j)] if i>0 and @array[i-1][j] == 0 # valid edge version 2
        end
      end
    end
  end

  # helper: takes indexes, returns node ident-number (0-80)
  def node_ident(i,j);  9*i+j;  end
end
